from collections import namedtuple
import sys
from typing import Union

import pytest

import bankid

if sys.version_info >= (3, 9):
    from builtins import type as Type
else:
    # Remove once PyBankID no longer support Python 3.8 or lower
    from typing import Type


@pytest.mark.parametrize(
    "exception_class,rfa",
    [
        (bankid.exceptions.AlreadyInProgressError, 4),
        (bankid.exceptions.InvalidParametersError, None),
        (bankid.exceptions.UnauthorizedError, None),
        (bankid.exceptions.NotFoundError, None),
        (bankid.exceptions.RequestTimeoutError, 5),
        (bankid.exceptions.InternalError, 5),
        (bankid.exceptions.MaintenanceError, 5),
        (bankid.exceptions.BankIDError, None),
    ],
)
def test_exceptions(exception_class: Type[Exception], rfa: Union[int, None]) -> None:
    e = exception_class()
    assert isinstance(e, bankid.exceptions.BankIDError)
    assert e.rfa == rfa


@pytest.mark.parametrize(
    "exception_class,error_code",
    [
        (bankid.exceptions.AlreadyInProgressError, "alreadyInProgress"),
        (bankid.exceptions.InvalidParametersError, "invalidParameters"),
        (bankid.exceptions.UnauthorizedError, "unauthorized"),
        (bankid.exceptions.NotFoundError, "notFound"),
        (bankid.exceptions.RequestTimeoutError, "requestTimeout"),
        (bankid.exceptions.InternalError, "internalError"),
        (bankid.exceptions.MaintenanceError, "maintenance"),
        (bankid.exceptions.BankIDError, "Unknown error code"),
    ],
)
def test_error_class_factory(exception_class: Type[Exception], error_code: str) -> None:
    MockResponse = namedtuple("MockResponse", ["json"])
    response = MockResponse(json=lambda: {"errorCode": error_code})
    # error: Argument 1 to "get_json_error_class" has incompatible type "MockResponse@41"; expected "Response"  [arg-type]
    e_class = bankid.exceptions.get_json_error_class(response)  # type: ignore[arg-type]
    assert isinstance(e_class, exception_class)



import random

import pytest
from typing import List, Tuple

from bankid.certs import get_test_cert_and_key


@pytest.fixture(scope="session")
def ip_address() -> str:
    return "127.0.0.1"


@pytest.fixture()
def cert_and_key() -> Tuple[str, str]:
    cert, key = get_test_cert_and_key()
    return str(cert), str(key)


@pytest.fixture()
def random_personal_number() -> str:
    """Simple random Swedish personal number generator."""

    def _luhn_digit(id_: str) -> int:
        """Calculate Luhn control digit for personal number.

        Code adapted from `Faker
        <https://github.com/joke2k/faker/blob/master/faker/providers/ssn
        /sv_SE/__init__.py>`_.

        :param id_: The partial number to calculate checksum of.
        :type id_: str
        :return: Integer digit in [0, 9].
        :rtype: int

        """

        def digits_of(n: int) -> List[int]:
            return [int(i) for i in str(n)]

        digits = digits_of(int(id_) * 10)
        checksum = sum(digits[-1::-2])
        for k in digits[-2::-2]:
            checksum += sum(digits_of(k * 2))
        checksum %= 10

        return checksum if checksum == 0 else 10 - checksum

    year = random.randint(1900, 2014)
    month = random.randint(1, 12)
    day = random.randint(1, 28)
    suffix = random.randint(0, 999)
    pn = "{0:04d}{1:02d}{2:02d}{3:03d}".format(year, month, day, suffix)
    return pn + str(_luhn_digit(pn[2:]))



# This directory contains the root certificates used by BankID.
# We have to pin these to prevent basic MITM attacks.

from pathlib import Path
from typing import Tuple


def get_test_cert_p12() -> Path:
    return (Path(__file__).parent / "FPTestcert4_20230629.p12").resolve()


def get_test_cert_and_key() -> Tuple[Path, Path]:
    return (
        (Path(__file__).parent / "FPTestcert4_20230629_cert.pem").resolve(),
        (Path(__file__).parent / "FPTestcert4_20230629_key.pem").resolve(),
    )



import base64
from datetime import datetime
from typing import Tuple, Dict, Any, Union, TypeVar, Generic
from urllib.parse import urljoin

from bankid.qr import generate_qr_code_content
from bankid.certutils import resolve_cert_path

import httpx

TClient = TypeVar("TClient", httpx.AsyncClient, httpx.Client)


class BankIDClientBaseclass(Generic[TClient]):
    """Baseclass for BankID clients.

    Both the synchronous and asynchronous clients inherit from this base class and has the methods implemented here.
    """

    client: TClient

    def __init__(
        self,
        certificates: Tuple[str, str],
        test_server: bool = False,
        request_timeout: int = 5,
    ):
        self.certs = certificates

        if test_server:
            self.api_url = "https://appapi2.test.bankid.com/rp/v6.0/"
            self.verify_cert = resolve_cert_path("appapi2.test.bankid.com.pem")
        else:
            self.api_url = "https://appapi2.bankid.com/rp/v6.0/"
            self.verify_cert = resolve_cert_path("appapi2.bankid.com.pem")

        self._auth_endpoint = urljoin(self.api_url, "auth")
        self._phone_auth_endpoint = urljoin(self.api_url, "phone/auth")
        self._sign_endpoint = urljoin(self.api_url, "sign")
        self._phone_sign_endpoint = urljoin(self.api_url, "phone/sign")
        self._collect_endpoint = urljoin(self.api_url, "collect")
        self._cancel_endpoint = urljoin(self.api_url, "cancel")

    @staticmethod
    def generate_qr_code_content(qr_start_token: str, start_t: Union[float, datetime], qr_start_secret: str) -> str:
        return generate_qr_code_content(qr_start_token, start_t, qr_start_secret)

    @staticmethod
    def _encode_user_data(user_data: str) -> str:
        return base64.b64encode(user_data.encode("utf-8")).decode("ascii")

    def _create_payload(
        self,
        end_user_ip: Union[str, None] = None,
        requirement: Union[Dict[str, Any], None] = None,
        user_visible_data: Union[str, None] = None,
        user_non_visible_data: Union[str, None] = None,
        user_visible_data_format: Union[str, None] = None,
    ) -> Dict[str, str]:
        data: Dict[str, Any] = {}
        if end_user_ip:
            data["endUserIp"] = end_user_ip
        if requirement and isinstance(requirement, dict):
            data["requirement"] = requirement
        if user_visible_data:
            data["userVisibleData"] = self._encode_user_data(user_visible_data)
        if user_non_visible_data:
            data["userNonVisibleData"] = self._encode_user_data(user_non_visible_data)
        if user_visible_data_format and user_visible_data_format == "simpleMarkdownV1":
            data["userVisibleDataFormat"] = "simpleMarkdownV1"
        return data



"""
:mod:`test_asyncclient`
=======================

.. module:: test_asyncclient
   :platform: Unix, Windows
   :synopsis:

.. moduleauthor:: tiwilliam <william@defunct.cc>

Created on 2023-12-15

"""

import uuid

import pytest
from typing import Tuple

from bankid import BankIDAsyncClient, exceptions


@pytest.mark.asyncio
async def test_authentication_and_collect(cert_and_key: Tuple[str, str], ip_address: str) -> None:
    """Authenticate call and then collect with the returned orderRef UUID."""
    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    assert "appapi2.test.bankid.com.pem" in str(c.verify_cert)
    out = await c.authenticate(ip_address)

    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = await c.collect(str(order_ref))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")


@pytest.mark.asyncio
async def test_sign_and_collect(cert_and_key: Tuple[str, str], ip_address: str) -> None:
    """Sign call and then collect with the returned orderRef UUID."""

    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    out = await c.sign(
        ip_address,
        user_visible_data="The data to be signed",
        user_non_visible_data="Non visible data",
    )
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = await c.collect(str(order_ref))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")


@pytest.mark.asyncio
async def test_phone_sign_and_collect(cert_and_key: Tuple[str, str], random_personal_number: str) -> None:
    """Phone sign call and then collect with the returned orderRef UUID."""

    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    out = await c.phone_sign(random_personal_number, "RP", user_visible_data="The data to be signed")
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = await c.collect(str(order_ref))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")


@pytest.mark.asyncio
async def test_invalid_orderref_raises_error(cert_and_key: Tuple[str, str]) -> None:
    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    with pytest.raises(exceptions.InvalidParametersError):
        await c.collect("invalid-uuid")


@pytest.mark.asyncio
async def test_already_in_progress_raises_error(cert_and_key: Tuple[str, str], ip_address: str, random_personal_number: str) -> None:
    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    await c.authenticate(ip_address, requirement={"personalNumber": random_personal_number})
    with pytest.raises(exceptions.AlreadyInProgressError):
        await c.authenticate(ip_address, requirement={"personalNumber": random_personal_number})


@pytest.mark.asyncio
async def test_already_in_progress_raises_error_2(cert_and_key: Tuple[str, str], ip_address: str, random_personal_number: str) -> None:
    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    await c.sign(
        ip_address,
        requirement={"personalNumber": random_personal_number},
        user_visible_data="Text to sign",
    )
    with pytest.raises(exceptions.AlreadyInProgressError):
        await c.sign(
            ip_address, requirement={"personalNumber": random_personal_number}, user_visible_data="Text to sign"
        )


@pytest.mark.asyncio
async def test_authentication_and_cancel(cert_and_key: Tuple[str, str], ip_address: str) -> None:
    """Authenticate call and then cancel it"""
    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    out = await c.authenticate(ip_address)
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = await c.collect(str(order_ref))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")
    success = await c.cancel(str(order_ref))
    assert success
    with pytest.raises(exceptions.InvalidParametersError):
        collect_status = await c.collect(str(order_ref))


@pytest.mark.asyncio
async def test_phone_authentication_and_cancel(cert_and_key: Tuple[str, str], random_personal_number: str) -> None:
    """Phone authenticate call and then cancel it"""

    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    out = await c.phone_authenticate(random_personal_number, "user")
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = await c.collect(str(order_ref))
    assert collect_status.get("status") == "pending"
    assert collect_status.get("hintCode") in ("outstandingTransaction", "noClient")
    success = await c.cancel(str(order_ref))
    assert success
    with pytest.raises(exceptions.InvalidParametersError):
        collect_status = await c.collect(str(order_ref))


@pytest.mark.asyncio
async def test_cancel_with_invalid_uuid(cert_and_key: Tuple[str, str]) -> None:
    c = BankIDAsyncClient(certificates=cert_and_key, test_server=True)
    invalid_order_ref = uuid.uuid4()
    with pytest.raises(exceptions.InvalidParametersError):
        await c.cancel(str(invalid_order_ref))



from typing import Any, Dict, Tuple, Union

import httpx

from bankid.baseclient import BankIDClientBaseclass
from bankid.exceptions import get_json_error_class


class BankIDAsyncClient(BankIDClientBaseclass[httpx.AsyncClient]):
    """The asynchronous client to use for communicating with BankID servers via the v6 API.

    :param certificates: Tuple of string paths to the certificate to use and
        the key to sign with.
    :type certificates: tuple
    :param test_server: Use the test server for authenticating and signing.
    :type test_server: bool
    :param request_timeout: Timeout for BankID requests.
    :type request_timeout: int

    """

    def __init__(self, certificates: Tuple[str, str], test_server: bool = False, request_timeout: int = 5):
        super().__init__(certificates, test_server, request_timeout)

        headers = {"Content-Type": "application/json"}
        self.client = httpx.AsyncClient(cert=self.certs, headers=headers, verify=str(self.verify_cert), timeout=request_timeout)

    async def authenticate(
        self,
        end_user_ip: str,
        requirement: Union[Dict[str, Any], None] = None,
        user_visible_data: Union[str, None]  = None,
        user_non_visible_data: Union[str, None]  = None,
        user_visible_data_format: Union[str, None]  = None,
    ) -> Dict[str, str]:
        """Request an authentication order. The :py:meth:`collect` method
        is used to query the status of the order.

        Example data returned:

        .. code-block:: json

            {
                "orderRef": "ee3421ea-2096-4000-8130-82648efe0927",
                "autoStartToken": "e8df5c3c-c67b-4a01-bfe5-fefeab760beb",
                "qrStartToken": "01f94e28-857f-4d8a-bf8e-6c5a24466658",
                "qrStartSecret": "b4214886-3b5b-46ab-bc08-6862fddc0e06"
            }

        :param end_user_ip: The user IP address as seen by RP. String. IPv4 and IPv6 is allowed.
        :type end_user_ip: str
        :param requirement: Requirements on how the auth order must be performed.
            See the section `Requirements <https://www.bankid.com/en/utvecklare/guider/teknisk-integrationsguide/graenssnittsbeskrivning/auth>`_ for more details.
        :type requirement: dict
        :param user_visible_data: Text displayed to the user during authentication with BankID,
            with the purpose of providing context for the authentication and to enable users
            to detect identification errors and averting fraud attempts.
        :type user_visible_data: str
        :param user_non_visible_data: Data is not displayed to the user.
        :type user_non_visible_data: str
        :param user_visible_data_format: If present, and set to “simpleMarkdownV1”,
            this parameter indicates that userVisibleData holds formatting characters which
            potentially make for a more pleasant user experience.
        :type user_visible_data_format: str
        :return: The order response.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        data = self._create_payload(
            end_user_ip,
            requirement=requirement,
            user_visible_data=user_visible_data,
            user_non_visible_data=user_non_visible_data,
            user_visible_data_format=user_visible_data_format,
        )

        response = await self.client.post(self._auth_endpoint, json=data)

        if response.status_code == 200:
            return response.json()  # type: ignore[no-any-return]
        else:
            raise get_json_error_class(response)

    async def phone_authenticate(
        self,
        personal_number: str,
        call_initiator: str,
        requirement: Union[Dict[str, Any], None] = None,
        user_visible_data: Union[str, None] = None,
        user_non_visible_data: Union[str, None] = None,
        user_visible_data_format: Union[str, None] = None,
    ) -> Dict[str, str]:
        """Initiates an authentication order when the user is talking
        to the RP over the phone. The :py:meth:`collect` method
        is used to query the status of the order.

        Example data returned:

        .. code-block:: json

            {
                "orderRef": "ee3421ea-2096-4000-8130-82648efe0927"
            }

        :param personal_number: The personal number of the user. 12 digits.
        :type personal_number: str
        :param call_initiator: Indicate if the user or the RP initiated the phone call.
            "user": user called the RP
            "RP": RP called the user
        :type call_initiator: str
        :param requirement: Requirements on how the auth order must be performed.
            See the section `Requirements <https://www.bankid.com/en/utvecklare/guider/teknisk-integrationsguide/graenssnittsbeskrivning/phone-auth>`_ for more details.
        :type requirement: dict
        :param user_visible_data: Text displayed to the user during authentication with BankID,
            with the purpose of providing context for the authentication and to enable users
            to detect identification errors and averting fraud attempts.
        :type user_visible_data: str
        :param user_non_visible_data: Data is not displayed to the user.
        :type user_non_visible_data: str
        :param user_visible_data_format: If present, and set to “simpleMarkdownV1”,
            this parameter indicates that userVisibleData holds formatting characters which
            potentially make for a more pleasant user experience.
        :type user_visible_data_format: str
        :return: The order response.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        if call_initiator not in ["user", "RP"]:
            raise ValueError("call_initiator must be either 'user' or 'RP'")

        data = self._create_payload(
            requirement=requirement,
            user_visible_data=user_visible_data,
            user_non_visible_data=user_non_visible_data,
            user_visible_data_format=user_visible_data_format,
        )
        data["personalNumber"] = personal_number
        data["callInitiator"] = call_initiator

        response = await self.client.post(self._phone_auth_endpoint, json=data)

        if response.status_code == 200:
            return response.json()  # type: ignore[no-any-return]
        else:
            raise get_json_error_class(response)

    async def sign(
        self,
        end_user_ip: str,
        user_visible_data: str,
        requirement: Union[Dict[str, Any], None] = None,
        user_non_visible_data: Union[str, None] = None,
        user_visible_data_format: Union[str, None] = None,
    ) -> Dict[str, str]:
        """Request a signing order. The :py:meth:`collect` method
        is used to query the status of the order.

        Example data returned:

        .. code-block:: json

            {
                "orderRef": "ee3421ea-2096-4000-8130-82648efe0927",
                "autoStartToken": "e8df5c3c-c67b-4a01-bfe5-fefeab760beb",
                "qrStartToken": "01f94e28-857f-4d8a-bf8e-6c5a24466658",
                "qrStartSecret": "b4214886-3b5b-46ab-bc08-6862fddc0e06"
            }

        :param end_user_ip: The user IP address as seen by RP. String. IPv4 and IPv6 is allowed.
        :type end_user_ip: str
        :param requirement: Requirements on how the sign order must be performed.
            See the section `Requirements <https://www.bankid.com/en/utvecklare/guider/teknisk-integrationsguide/graenssnittsbeskrivning/sign>`_ for more details.
        :type requirement: dict
        :param user_visible_data: Text to be displayed to the user.
        :type user_visible_data: str
        :param user_non_visible_data: Data is not displayed to the user.
        :type user_non_visible_data: str
        :param user_visible_data_format: If present, and set to “simpleMarkdownV1”,
            this parameter indicates that userVisibleData holds formatting characters which
            potentially make for a more pleasant user experience.
        :type user_visible_data_format: str
        :return: The order response.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                     when error has been returned from server.

        """
        data = self._create_payload(
            end_user_ip,
            requirement=requirement,
            user_visible_data=user_visible_data,
            user_non_visible_data=user_non_visible_data,
            user_visible_data_format=user_visible_data_format,
        )

        response = await self.client.post(self._sign_endpoint, json=data)

        if response.status_code == 200:
            return response.json()  # type: ignore[no-any-return]
        else:
            raise get_json_error_class(response)

    async def phone_sign(
        self,
        personal_number: str,
        call_initiator: str,
        user_visible_data: str,
        requirement: Union[Dict[str, Any], None] = None,
        user_non_visible_data: Union[str, None] = None,
        user_visible_data_format: Union[str, None] = None,
    ) -> Dict[str, str]:
        """Initiates an authentication order when the user is talking to
        the RP over the phone. The :py:meth:`collect` method
        is used to query the status of the order.

        Example data returned:

        .. code-block:: json

            {
                "orderRef": "ee3421ea-2096-4000-8130-82648efe0927"
            }

        :param personal_number: The personal number of the user. 12 digits.
        :type personal_number: str
        :param call_initiator: Indicate if the user or the RP initiated the phone call.
            "user": user called the RP
            "RP": RP called the user
        :type call_initiator: str
        :param requirement: Requirements on how the sign order must be performed.
            See the section `Requirements <https://www.bankid.com/en/utvecklare/guider/teknisk-integrationsguide/graenssnittsbeskrivning/sign>`_ for more details.
        :type requirement: dict
        :param user_visible_data: Text to be displayed to the user.
        :type user_visible_data: str
        :param user_non_visible_data: Data is not displayed to the user.
        :type user_non_visible_data: str
        :param user_visible_data_format: If present, and set to “simpleMarkdownV1”,
            this parameter indicates that userVisibleData holds formatting characters which
            potentially make for a more pleasant user experience.
        :type user_visible_data_format: str
        :return: The order response.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                     when error has been returned from server.

        """
        if call_initiator not in ["user", "RP"]:
            raise ValueError("call_initiator must be either 'user' or 'RP'")

        data = self._create_payload(
            requirement=requirement,
            user_visible_data=user_visible_data,
            user_non_visible_data=user_non_visible_data,
            user_visible_data_format=user_visible_data_format,
        )
        data["personalNumber"] = personal_number
        data["callInitiator"] = call_initiator

        response = await self.client.post(self._phone_sign_endpoint, json=data)

        if response.status_code == 200:
            return response.json()  # type: ignore[no-any-return]
        else:
            raise get_json_error_class(response)

    async def collect(self, order_ref: str) -> dict:
        """Collects the result of a sign or auth order using the
        ``orderRef`` as reference.

        RP should keep on calling collect every two seconds if status is pending.
        RP must abort if status indicates failed. The user identity is returned
        when complete.

        Example collect results returned while authentication or signing is
        still pending:

        .. code-block:: json

            {
                "orderRef":"131daac9-16c6-4618-beb0-365768f37288",
                "status":"pending",
                "hintCode":"userSign"
            }

        Example collect result when authentication or signing has failed:

        .. code-block:: json

            {
                "orderRef":"131daac9-16c6-4618-beb0-365768f37288",
                "status":"failed",
                "hintCode":"userCancel"
            }

        Example collect result when authentication or signing is successful
        and completed:

        .. code-block:: json

            {
                "orderRef": "131daac9-16c6-4618-beb0-365768f37288",
                "status": "complete",
                "completionData": {
                    "user": {
                        "personalNumber": "190000000000",
                        "name": "Karl Karlsson",
                        "givenName": "Karl",
                        "surname": "Karlsson"
                    },
                    "device": {
                        "ipAddress": "192.168.0.1"
                    },
                    "bankIdIssueDate": "2020-02-01",
                    "signature": "<base64-encoded data>",
                    "ocspResponse": "<base64-encoded data>"
                }
            }

        See `BankID Integration Guide <https://www.bankid.com/en/utvecklare/guider/teknisk-integrationsguide/graenssnittsbeskrivning/collect>`_
        for more details about how to inform end user of the current status,
        whether it is pending, failed or completed.

        :param order_ref: The ``orderRef`` UUID returned from auth or sign.
        :type order_ref: str
        :return: The CollectResponse parsed to a dictionary.
        :rtype: dict
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        response = await self.client.post(self._collect_endpoint, json={"orderRef": order_ref})

        if response.status_code == 200:
            return response.json()  # type: ignore[no-any-return]
        else:
            raise get_json_error_class(response)

    async def cancel(self, order_ref: str) -> bool:
        """Cancels an ongoing sign or auth order.

        This is typically used if the user cancels the order
        in your service or app.

        :param order_ref: The UUID string specifying which order to cancel.
        :type order_ref: str
        :return: Boolean regarding success of cancellation.
        :rtype: bool
        :raises BankIDError: raises a subclass of this error
                             when error has been returned from server.

        """
        response = await self.client.post(self._cancel_endpoint, json={"orderRef": order_ref})

        if response.status_code == 200:
            return response.json() == {}  # type: ignore[no-any-return]
        else:
            raise get_json_error_class(response)



import os

from pytest import TempdirFactory

import bankid


def test_create_bankid_test_server_cert_and_key(tmpdir_factory: TempdirFactory) -> None:
    paths = bankid.certutils.create_bankid_test_server_cert_and_key(tmpdir_factory.mktemp("certs"))
    assert os.path.exists(paths[0])
    assert os.path.exists(paths[1])
    try:
        os.remove(paths[0])
        os.remove(paths[1])
    except Exception:
        pass



import pathlib
import time
import uuid

from flask import Flask, make_response, render_template, request, jsonify
from flask_caching import Cache
from bankid import BankIDClient
from bankid.exceptions import BankIDError
from bankid.certutils import create_bankid_test_server_cert_and_key

USE_TEST_SERVER = True

app = Flask(__name__)
cache = Cache(app, config={"CACHE_TYPE": "SimpleCache"})

# The client should be initialized in a better way, e.g. with Flask_BankID so that it is stored in the
# Flask app. For this demo it is sufficient to let it reside globally in this file.
if USE_TEST_SERVER:
    cert_paths = create_bankid_test_server_cert_and_key(str(pathlib.Path(__file__).parent))
    client = BankIDClient(cert_paths, test_server=True)
else:
    # Set your own cert paths for you production certificate and key here.
    # Note that my recommendation is to get it to work with
    # test server certs first!
    cert_paths = ("certificate.pem", "key.pem")
    client = BankIDClient(cert_paths, test_server=False)


# Frontend pages


@app.route("/")
def index():
    """Landing page, with form to fill in Personal Identity Number"""
    return render_template("index.html")


@app.route("/auth-complete")
def auth_complete():
    """When the authentication is either completed or failed, this is where the user ends up"""
    # Try to get the cookie set from the collect method,
    auth_cookie = request.cookies.get("QRDemo-Auth")
    if auth_cookie:
        # There was a cookie. Get the collect response stored in the cache.
        collect_response = cache.get(uuid.UUID(auth_cookie))
    else:
        # No cookie. Authentication failed.
        collect_response = {}
        age = None

    response = make_response(render_template("auth_complete.html", auth=collect_response))
    # Unsetting the cookie to make the app capable of being run again.
    response.set_cookie("QRDemo-Auth", "", expires=0)
    return response


# Backend API methods


@app.route("/send-initiate", methods=["POST"])
def initiate():
    """Initiate a BankID Authentication session and cache details needed for QR code generation"""
    # Note that empty personal number is allowed here! That means that the
    pn = request.form.get("personnummer")

    # By knowing the personal number of the one you want to autenticate, send in the
    # personal number as a requirement to ensure that only autentication with that personal
    # number is allowed.

    # Make Auth call to BankID.
    resp = client.authenticate(
        end_user_ip=request.remote_addr,  # Get the IP of the device making the request.
        requirement={"personalNumber": pn} if pn else None,  # Set to True if PN is provided. Recommended.
    )
    # Record when this response was received. This is needed for generating sequential, animated QR codes.
    resp["start_t"] = time.time()
    # Store these details for later use in some way. Here I use a very simple memory cache; not recommended for
    # multi-instance apps. Using orderRef as key since it is unique and can be sent in a GET URL without problem.
    cache.set(resp.get("orderRef"), resp, timeout=5 * 60)
    # Generate the first QR code to display to user.
    qr_content_0 = client.generate_qr_code_content(resp["qrStartToken"], resp["start_t"], resp["qrStartSecret"])
    return render_template(
        "qr.html",
        order_ref=resp["orderRef"],
        auto_start_token=resp["autoStartToken"],
        qr_content=qr_content_0,
    )


@app.route("/get-qr-code/<order_ref>")
def get_qr_code(order_ref: str):
    """Get the current QR code content to generate QR code from"""
    x = cache.get(order_ref)
    if x is None:
        qr_content = ""
    else:
        qr_content = client.generate_qr_code_content(x["qrStartToken"], x["start_t"], x["qrStartSecret"])
    response = make_response(qr_content, 200)
    response.mimetype = "text/plain"
    return response


@app.route("/collect/<order_ref>")
def collect(order_ref: str):
    """Make collect calls to the BankID servers"""
    try:
        collect_response = client.collect(order_ref)
    except BankIDError as e:
        return jsonify(e.json), 400

    if collect_response.get("status") == "complete":
        # Create or Login the newly authenticated user, give it a session token or something similar.
        # Here I will use an ugly hack and just stick a generated UUID in a cookie so the frontend can tell
        # the next page that it is successfully signed in. The collect response is then stored in the cache using
        # the same UUID so that it can be recovered later.
        # Note that this is NOT production quality code, merely for proving a point.
        session_id = uuid.uuid4()
        cache.set(session_id, collect_response)
        response = make_response(collect_response, 200)
        response.mimetype = "application/json"
        response.set_cookie("QRDemo-Auth", str(session_id))
        return response
    else:
        return jsonify(collect_response)


@app.route("/cancel/<order_ref>")
def cancel(order_ref: str):
    """Make a cancel call to the BankID servers"""
    cancel_response = client.cancel(order_ref)
    if cancel_response:
        cache.delete(order_ref)
        response = make_response(str(cancel_response), 200)
        response.mimetype = "text/plain"
        response.delete_cookie("QRDemo-Auth")
        return response
    else:
        cache.delete(order_ref)
        response = make_response(str(cancel_response), 500)
        response.mimetype = "text/plain"
        return response


@app.errorhandler(500)
def internal_error(error):
    return str(error)



# -*- coding: utf-8 -*-
"""
:mod:`bankid`
=============

Created 2016-11-28 by hbldh

PyBankID is a client for providing BankID services as a Relying Party, i.e.
providing authentication and signing functionality to end users. This package
provides a simplifying interface for initiating authentication
and signing orders and then collecting the results from the BankID servers.

If you intend to use PyBankID in your project, you are advised to read
the `BankID Relying Party Integration Guide
<https://www.bankid.com/en/utvecklare/guider/teknisk-integrationsguide>`_ before
doing anything else. There, one can find information
about how the BankID methods are defined and how to use them.

"""

from bankid import exceptions
from bankid.__version__ import __version__, version
from bankid.certutils import create_bankid_test_server_cert_and_key
from bankid.syncclient import BankIDClient
from bankid.asyncclient import BankIDAsyncClient
from bankid.qr import generate_qr_code_content

__all__ = [
    "BankIDClient",
    "BankIDAsyncClient",
    "exceptions",
    "create_bankid_test_server_cert_and_key",
    "generate_qr_code_content",
    "__version__",
    "version",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`test_syncclient`
======================

.. module:: test_syncclient
   :platform: Unix, Windows
   :synopsis:

.. moduleauthor:: mxamin <amin.solhizadeh@gmail.com>

Created on 2024-01-18

"""
import uuid

import pytest
from typing import Tuple

try:
    from unittest import mock
except:
    import mock  # type: ignore[no-redef]

from bankid import BankIDClient, exceptions


def test_authentication_and_collect(cert_and_key: Tuple[str, str], ip_address: str, random_personal_number: str) -> None:
    """Authenticate call and then collect with the returned orderRef UUID."""

    c = BankIDClient(certificates=cert_and_key, test_server=True)
    assert "appapi2.test.bankid.com.pem" in str(c.verify_cert)
    out = c.authenticate(ip_address, requirement={"personalNumber": random_personal_number})
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = c.collect(str(order_ref))
    assert collect_status["status"] == "pending"
    assert collect_status["hintCode"] in ("outstandingTransaction", "noClient")


def test_sign_and_collect(cert_and_key: Tuple[str, str], ip_address: str) -> None:
    """Sign call and then collect with the returned orderRef UUID."""

    c = BankIDClient(certificates=cert_and_key, test_server=True)
    out = c.sign(
        ip_address,
        user_visible_data="The data to be signed",
        user_non_visible_data="Non visible data",
    )
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = c.collect(str(order_ref))
    assert collect_status["status"] == "pending"
    assert collect_status["hintCode"] in ("outstandingTransaction", "noClient")


def test_phone_sign_and_collect(cert_and_key: Tuple[str, str], random_personal_number: str) -> None:
    """Phone sign call and then collect with the returned orderRef UUID."""

    c = BankIDClient(certificates=cert_and_key, test_server=True)
    out = c.phone_sign(random_personal_number, "user", user_visible_data="The data to be signed")
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = c.collect(str(order_ref))
    assert collect_status["status"] == "pending"
    assert collect_status["hintCode"] in ("outstandingTransaction", "noClient")


def test_invalid_orderref_raises_error(cert_and_key: Tuple[str, str]) -> None:
    c = BankIDClient(certificates=cert_and_key, test_server=True)
    with pytest.raises(exceptions.InvalidParametersError):
        collect_status = c.collect("invalid-uuid")


def test_already_in_progress_raises_error(cert_and_key: Tuple[str, str], ip_address: str, random_personal_number: str) -> None:
    c = BankIDClient(certificates=cert_and_key, test_server=True)
    out = c.authenticate(ip_address, requirement={"personalNumber": random_personal_number})
    with pytest.raises(exceptions.AlreadyInProgressError):
        out2 = c.authenticate(ip_address, requirement={"personalNumber": random_personal_number})


def test_already_in_progress_raises_error_2(cert_and_key: Tuple[str, str], ip_address: str, random_personal_number: str) -> None:
    c = BankIDClient(certificates=cert_and_key, test_server=True)
    out = c.sign(ip_address, requirement={"personalNumber": random_personal_number}, user_visible_data="Text to sign")
    with pytest.raises(exceptions.AlreadyInProgressError):
        out2 = c.sign(
            ip_address, requirement={"personalNumber": random_personal_number}, user_visible_data="Text to sign"
        )


def test_authentication_and_cancel(cert_and_key: Tuple[str, str], ip_address: str, random_personal_number: str) -> None:
    """Authenticate call and then cancel it"""

    c = BankIDClient(certificates=cert_and_key, test_server=True)
    out = c.authenticate(ip_address, requirement={"personalNumber": random_personal_number})
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = c.collect(str(order_ref))
    assert collect_status["status"] == "pending"
    assert collect_status["hintCode"] in ("outstandingTransaction", "noClient")
    success = c.cancel(str(order_ref))
    assert success
    with pytest.raises(exceptions.InvalidParametersError):
        collect_status = c.collect(str(order_ref))


def test_phone_authentication_and_cancel(cert_and_key: Tuple[str, str], random_personal_number: str) -> None:
    """Phone authenticate call and then cancel it"""

    c = BankIDClient(certificates=cert_and_key, test_server=True)
    out = c.phone_authenticate(random_personal_number, "user")
    assert isinstance(out, dict)
    # UUID.__init__ performs the UUID compliance assertion.
    order_ref = uuid.UUID(out["orderRef"], version=4)
    collect_status = c.collect(str(order_ref))
    assert collect_status["status"] == "pending"
    assert collect_status["hintCode"] in ("outstandingTransaction", "noClient")
    success = c.cancel(str(order_ref))
    assert success
    with pytest.raises(exceptions.InvalidParametersError):
        collect_status = c.collect(str(order_ref))


def test_cancel_with_invalid_uuid(cert_and_key: Tuple[str, str]) -> None:
    c = BankIDClient(certificates=cert_and_key, test_server=True)
    invalid_order_ref = uuid.uuid4()
    with pytest.raises(exceptions.InvalidParametersError):
        cancel_status = c.cancel(str(invalid_order_ref))


@pytest.mark.parametrize(
    "test_server, endpoint",
    [(False, "appapi2.bankid.com"), (True, "appapi2.test.bankid.com")],
)
def test_correct_prod_server_urls(cert_and_key: Tuple[str, str], test_server: bool, endpoint: str) -> None:
    c = BankIDClient(certificates=cert_and_key, test_server=test_server)
    assert c.api_url == "https://{0}/rp/v6.0/".format(endpoint)
    assert "{0}.pem".format(endpoint) in str(c.verify_cert)



from typing import Union

import hashlib
import hmac
import time
from datetime import datetime
from math import floor


def generate_qr_code_content(qr_start_token: str, start_t: Union[float, datetime], qr_start_secret: str) -> str:
    """Given QR start token, time.time() or UTC datetime when initiated authentication call was made and the
    QR start secret, calculate the current QR code content to display.
    """
    if isinstance(start_t, datetime):
        start_t = start_t.timestamp()
    elapsed_seconds_since_call = int(floor(time.time() - start_t))
    qr_auth_code = hmac.new(
        qr_start_secret.encode(),
        msg=str(elapsed_seconds_since_call).encode(),
        digestmod=hashlib.sha256,
    ).hexdigest()
    return f"bankid.{qr_start_token}.{elapsed_seconds_since_call}.{qr_auth_code}"
